#ifndef DISPLAYIMAGE_H
#define DISPLAYIMAGE_H

#include "../image/image.h"
#include "../point/point.h"
#include <gtkmm/window.h>
#include <gtkmm/box.h>
#include <gtkmm/image.h>
#include <gtkmm/scrolledwindow.h>
#include <gtkmm/adjustment.h>
#include <gtkmm/label.h>
#include <gtkmm/actiongroup.h>
#include <gtkmm/uimanager.h>
#include <gtkmm/eventbox.h>
#include <gdk/gdkevents.h>
#include <gdkmm/pixbuf.h>
#include <gtkmm/filechooserdialog.h>
#include <gtkmm/drawingarea.h>

using namespace std;

namespace CommonTilton
{

  class DisplayImage : public Gtk::Window
  {
   public:
      DisplayImage();
      virtual ~DisplayImage();

      static void set_static_values(const Image& input_image);

    // For RGB Image and Region Mean Image (type 1)
      void init_rgb(Image& input_image, const Glib::ustring& title);
      void init_rgb(Image& input_image, Image& mask_image, const Glib::ustring& title);
      void reinit_rgb();
    // For Segmentation Classes Slice and Segmentation Objects Slice (type 2)
      void init_seg(Image& input_image, const Glib::ustring& title);
      void reinit_seg();
    // For Current Class Labels (type 3)
      void init_region_label(Image& input_image, Image& mask_image, const Glib::ustring& title);
      void reinit_region_label();
    // For Hierarchical Boundary Map (type 4)
      void init_boundary(Image& input_image, const Glib::ustring& title);
      void reinit_boundary();
    // For Reference Images - (type 5) (Like type 6, but no scaling or masking)
      void init_reference(Image& input_image, const Glib::ustring& title);
    // For General Float or Double Images (e.g. Std. Dev. Maps, Boundary Pixel Ratio (BPRatio) Maps,
    // NDVI Images, NDSI Images, etc. - type 6)
      void init_float(Image& input_image, const Glib::ustring& title);
      void init_float(Image& input_image, Image& mask_image, const Glib::ustring& title);
      void reinit_float();
    // For MODIS Snow Maps (type 7)
      void init_snow(Image& input_image, const Glib::ustring& title);
    // For MODIS Thin Cloud Mask (type 8)
      void init_thin_cloud(Image& input_image, const Glib::ustring& title);
    // For Revised Snow Map (type 9)
      void init_revised_snow(Image& input_image, const Glib::ustring& title);
      void reinit_revised_snow();
    // For Drawing a Path (type 10)
      void init_path(Image& input_image, vector<Point>& input_path, const Glib::ustring& title);
      void reinit_path(vector<Point>& input_path);

    // Other public functions
      void copy_display(Image& copy_image);
      void revise_display(Image& revise_image);
      void revise_display(int& label);
      void link_image(DisplayImage *image);
      void set_click_location(int& col, int& row);
      bool get_click_location(int& col, int& row);
      void set_label_mask(const unsigned int& selLabel);
      bool set_display_highlight(const unsigned int& selLabel);
      bool set_display_highlight(const unsigned int& selLabel, Image& labelImage);
      bool set_display_highlight(const unsigned int& selLabel, Image* labelImage);
      bool set_display_highlight(const float& threshold, float* region_mean);
      bool get_display_highlight(const int& col, const int& row);
      void reset_display_highlight();
      void highlight_region(int& colSelect, int& rowSelect);
      void highlight_region(Image* sourceImage, int& colSelect, int& rowSelect);
      void highlight_select_region(int& colSelect, int& rowSelect);
      void highlight_circle_roi(Cairo::RefPtr<Cairo::Context>& display_Context);
      bool set_focus(const bool& focusAll);
      bool set_linked_focus(const bool& focusAll);
      void display_resize(const int& width, const int& height);
      double get_X_zoom_factor() const { return X_zoom_factor; }
      double get_Y_zoom_factor() const { return Y_zoom_factor; }
      void registered_data_copy(Image& destImage, const int& band)
           { destImage.registered_data_copy(band,displayImage,band); displayImage.close(); return; }
      void remove_display_file();
      void set_button_press_monitor_flag(const bool& value);

      //Signal accessors:
      typedef sigc::signal<void> type_signal_button_press_event;
      type_signal_button_press_event signal_button_press_event();

    protected:
    //Signal handlers:
      virtual void on_menu_save_png();
      virtual void on_menu_save_data();
      virtual void on_menu_zoom_in();
      virtual void on_menu_zoom_out();
      virtual void on_menu_select_region();
      virtual void on_menu_circle_roi();
      virtual void on_menu_help();
      virtual void on_menu_file_close();
      void on_window_check_resize();
      bool on_thumb_motion_notify_event(GdkEventMotion* event);
      bool on_thumb_button_press_event(GdkEventButton* event);
      bool on_display_expose_event(GdkEventExpose* event);
      bool on_display_motion_notify_event(GdkEventMotion* event);
      bool on_display_button_press_event(GdkEventButton* event);
      bool on_display_button_release_event(GdkEventButton* event);
      void on_hadjustment_value_changed();
      void on_vadjustment_value_changed();

    //Signal accessors:
      type_signal_button_press_event m_signal_button_press_event;

    //Child widgets:
      Gtk::VBox m_Box;
      Gtk::EventBox thumb_Event;
      Gtk::Image thumb_Image;
      Gtk::DrawingArea display_DrawingArea;
      Gtk::EventBox display_Event;
      Gtk::Image display_Image;
      Gtk::Adjustment hadjustment, vadjustment;
      Gtk::ScrolledWindow scrolled_Window;
      Gtk::Label pixel_Label;

      Glib::RefPtr<Gtk::UIManager> m_refUIManager;
      Glib::RefPtr<Gtk::ActionGroup> m_refActionGroup;

      Glib::RefPtr<Gdk::Pixbuf> fullPixbuf, thumbPixbuf, displayPixbuf;

    private:
      bool set_display_file();
      void calc_defaults();
      void init();
      void display_thumb_image_set();
      void display_thumb_image_reset();
      void display_zoom_factor_reset(const double& factor);
      void set_pixel_label();
      int propagate(Image* sourceImage, int searchValue, 
                    int minSearchCol, int maxSearchCol, int minSearchRow, int maxSearchRow);
      int propagate(int minSearchCol, int maxSearchCol, int minSearchRow, int maxSearchRow);
      void display_refresh();
      void display_refresh(vector<Point>& input_path);
      void draw_path(Cairo::RefPtr<Cairo::Context>& display_Context);
      double hadjustment_value, vadjustment_value;
      string display_file;
      Image displayImage, *inputImage, *maskImage;
      RHSEGDType dtype;
      bool *display_highlight, *temp_highlight;
      unsigned char display_type, *display_data;
      unsigned int  numberLinked;
      double X_zoom_factor, Y_zoom_factor, zoom_factor, thumb_factor;
      int colPosition, rowPosition, colClick, rowClick;
      bool clickFlag, resizeFlag;
      bool mask_valid, circle_roi_flag, select_region_flag, button_pressed_flag, button_press_monitor_flag;
      int ncols, nrows, display_ncols, display_nrows, thumb_ncols, thumb_nrows;
      int window_ncols, window_nrows, visible_ncols, visible_nrows;
      DisplayImage *linkedImages[MAX_LINKED_IMAGES];
      double scale, red_scale, green_scale, blue_scale;
      double offset, red_offset, green_offset, blue_offset;
      vector<Point> path;
      static int base_ncols, base_nrows;
      static double base_X_gsd, base_Y_gsd;
  };

} // CommonTilton

#endif /* DISPLAYIMAGE_H */
