#ifndef IMAGE_H
#define IMAGE_H

#include <string>
#include <vector>
#include <iostream>
#include <fstream>
#ifdef GDAL
#include <gdal_priv.h>
#include <gdalwarper.h>
#endif // GDAL

using namespace std;

namespace CommonTilton
{

  enum RHSEGDType { Unknown, UInt8, UInt16, UInt32, Float32 };

  struct ColorMap
  {
    unsigned char  r,g,b;
    unsigned short sum; 
  };

  // Comparison criterion for heap
  class ColorLessThan
  {
    public:
      bool operator ( ) (const ColorMap& colormap1, const ColorMap& colormap2)
      {
        if ((colormap1.sum == colormap2.sum) && (colormap1.b == colormap2.b) && 
            (colormap1.g == colormap2.g))
          return (colormap1.r < colormap2.r);
        else if ((colormap1.sum == colormap2.sum) && (colormap1.b == colormap2.b))
          return (colormap1.g < colormap2.g);
        else if (colormap1.sum == colormap2.sum)
          return (colormap1.r < colormap2.r);
        else
          return (colormap1.sum < colormap2.sum);
      }
  };

  class Image 
  {
    public:
    // Constructors and Destructor
      Image();
#ifdef GDAL
      Image(const string& file_name);
#endif
      virtual ~Image();

    // Member functions
      bool gdal_valid() const { return gdal_flag; }
      bool info_valid() const { return info_flag; }
#ifdef GDAL
      GDALDataset *get_imageDataset() const { return imageDataset; }
      bool read_info(const string& file_name);
      bool get_info();
      void registered_copy(const string& file_name, const Image& baseImage, Image& sourceImage);
      void copy(const string& file_name, const Image& baseImage);
      void formatted_copy(const string& file_name, const Image& baseImage, const string& source_driver_description);
      bool create_mask(const string& file_name, Image& source_image, const string& source_driver_description);
      bool create_geo_copy(const string& file_name, const Image& baseImage, const int& copy_nbands, 
                           const GDALDataType& copy_data_type, const string& copy_driver_description);
      bool create_geo_subset_copy(const string& file_name, const Image& baseImage,
                                  const Image& subset_image, const int& copy_nbands,
                                  const GDALDataType& copy_data_type, const string& copy_driver_description);
      void create(const string& file_name, const int& n_cols, const int& n_rows,
                  const int& n_bands, const GDALDataType& data_type,
                  const double& Xoffset, const double& Yoffset, const double& Xgsd, const double& Ygsd, 
                  const string& n_driver_description);
      void create(const string& file_name, const int& n_cols, const int& n_rows,
                  const int& n_bands, const GDALDataType& data_type,
                  const string& n_driver_description);
      void set_snow_color_table();
#else
      bool create_geo_copy(const string& file_name, const Image& baseImage, const int& copy_nbands, 
                           RHSEGDType& copy_dtype);
      bool create_geo_subset_copy(const string& file_name, const Image& baseImage,
                                  const Image& subset_image, const int& copy_nbands, RHSEGDType& copy_dtype);
#endif
      void create(const string& file_name, const int& n_cols, const int& n_rows,
                  const int& n_bands, const RHSEGDType& d_type,
                  const double& Xoffset, const double& Yoffset, const double& Xgsd, const double& Ygsd,
                  const bool& w_flag);
#ifdef THREEDIM
      void create(const string& file_name, const int& n_cols, const int& n_rows, const int& n_slices,
                  const int& n_bands, const RHSEGDType& d_type, const bool& w_flag);
#endif
      void create(const string& file_name, const int& n_cols, const int& n_rows,
                  const int& n_bands, const RHSEGDType& d_type, const bool& w_flag);
      void set_geotransform(const double& Xoffset, const double& Yoffset, const double& Xgsd, const double& Ygsd);
      void set_no_data_value(const double& value);
      void set_rgb_display_bands(const int& red_band, const int& green_band, 
                                 const int& blue_band);

      string get_file_name() const { return image_file; }
      int get_ncols() const { return ncols; }
      int get_nrows() const { return nrows; }
      int get_nslices() const { return nslices; }
      int get_nbands() const { return nbands; }
      int get_red_display_band() const { return red_display_band; }
      int get_green_display_band() const { return green_display_band; }
      int get_blue_display_band() const { return blue_display_band; }
      RHSEGDType get_dtype() const { return dtype; }
#ifdef GDAL
      GDALDataType get_data_type() const { return data_type; }
      string get_driver_description() const { return driver_description; }
#endif

      bool geotransform_valid() const { return geotransform_flag; }
      bool data_valid() const { return data_flag; }

      bool clip_and_mask(const double& clip_min, const double& clip_max, const double& no_data_val);
      bool no_data_value_valid(const int& band) const { return no_data_value_flag[band]; }
      double get_no_data_value(const int& band) { return no_data_value[band]; }

      void computeMinMax();
      void computeMinMax(Image& maskImage);
      void resetMinMax();
      void resetMinMax(Image& maskImage);
      double getMinimum(const int& band);
      double getMinimum(const int& band, Image& maskImage);
      double getMaximum(const int& band);
      double getMaximum(const int& band, Image& maskImage);

      bool stretch(const int& band, Image& sourceImage, const int& nb_levels);
      bool stretch(const int& band, Image& sourceImage, const int& nb_levels, Image& maskImage);
      void computeHistoEqMap(const int& band, const int& nb_levels);
      void computeHistoEqMap(const int& band, const int& nb_levels, Image& maskImage);
      bool performHistoEq(const int& band, Image& sourceImage);
      bool performHistoEq(const int& band, Image& sourceImage, Image& maskImage);
      bool histo_eq_map_valid(const int& band) const { if (histo_eq_map_flag.empty())
                                                         return false;
                                                       else
                                                         return histo_eq_map_flag[band]; }
      double get_histo_lookup(const double& lookup_value, const int& band);
      void copy_histo_lookup(const int& band, const Image& sourceImage, const int& sourceBand);
                 
      void resize_colormap(const unsigned int& colormap_size);
      void set_colormap_value(const unsigned int &index, const unsigned int& red_value,
                              const unsigned int& green_value, const unsigned int& blue_value);
      void compute_colormap(const unsigned int& colormap_size);
      unsigned char get_red_colormap(const unsigned int& data_value);
      unsigned char get_green_colormap(const unsigned int& data_value);
      unsigned char get_blue_colormap(const unsigned int& data_value);
      void print_colormap();
      void scale_offset(const int& new_min, const int& new_max, const int& band, Image& maskImage);
      void scale_offset_SI(Image& SIImage, const float SI_scale, const float SI_offset, const int& SI_band,
                           const int& min_valid, const int& max_valid);
      void chi_square_ratio(Image& posChiSquareImage, Image& negChiSquareImage, const double& chi_square_threshold,
                            Image& dofImage, const int& dof_threshold);
#ifdef MODIS
      void copy_swath_lat_long_data(Image& swathLatitudeImage, Image& swathLongitudeImage);
      void copy_proj_lat_long_data(Image& projLatitudeImage, Image& projLongitudeImage);
      void NNProjColRow(Image& swathLatLongImage, Image& projLatLongImage);
      void NNProj(Image& swathDataImage, Image& projColRowImage);
      void modis_scale_and_correct(Image& inputImage, const int& input_band, const float& scale, const float& offset,
                                   Image& solarZenithImage, const int& output_band);
      void compute_surface_temperature(Image& inputImage, vector<float>& scale, vector<float>& offset, 
                                       const int& band_31, const int& band_32);
      void compute_thin_cirrus_cloud_mask(Image& cloudMaskImage, Image& cloudMaskQAImage, Image& snowCoverImage);
#endif
      void upsample_copy(Image& sourceImage, const double& continuity_ratio);
      void compute_ndsi(Image& inputImage, const int& green_band, const int& MIR_band, Image& maskImage);
      void compute_ndsi(Image& inputImage, const int& green_band, const int& MIR_band);

      void compute_ndvi(Image& inputImage, const int& NIR_band, const int& red_band, Image& maskImage);
      void compute_ndvi(Image& inputImage, const int& NIR_band, const int& red_band);
      void compute_difference(Image& firstImage, Image& secondImage);
      void compare(Image& firstImage, Image& secondImage);
      void compare(Image& firstImage, Image& secondImage, const double& UTM_X_shift, const double& UTM_Y_shift);
      void compare(Image& compareImage, const string& compare_type, const double& compare_value);

#ifdef THREEDIM
      bool data_valid(const int& col, const int& row, const int& slice, const int& band);
#endif
      bool data_valid(const int& col, const int& row, const int& band);
      bool data_valid(const double& UTM_X, const double& UTM_Y, const int& band);
      double offset_for_rounding(const double& value);
#ifdef THREEDIM
      double get_data(const int& col, const int& row, const int& slice, const int& band);
#endif
      double get_data(const int& col, const int& row, const int& band);
      double get_data(const double& UTM_X, const double& UTM_Y, const int& band);
      double get_X_offset() const { return X_offset; }
      double get_Y_offset() const { return Y_offset; }
      double get_X_gsd() const { return X_gsd; }
      double get_Y_gsd() const { return Y_gsd; }
      double get_UTM_X(const int& col) const { return (X_offset + col*X_gsd); }
      double get_UTM_Y(const int& row) const { return (Y_offset + row*Y_gsd); }

#ifdef THREEDIM
      void put_data(const double& value, const int& col, const int& row, const int& slice, const int& band);
#endif
      void put_data(const double& value, const int& col, const int& row, const int& band);
      void put_data(const double& value, const double& UTM_X, const double& UTM_Y, const int& band);
      void put_data_values(const double& value);
#ifdef GDAL
      void put_no_data_value();
      bool update_mask(Image& sourceImage);
#endif
      void increment(const int& col, const int& row, const int& band);
      void registered_data_copy(const int& band, Image& sourceImage, const int& source_band);
      void registered_data_copy(const int& band, Image& sourceImage, const int& source_band, Image& maskImage);
/*
      void forward_data_copy(const int& band, Image& forwardProjRowColImage, Image& sourceImage, const int& source_band);
      void reverse_data_copy(const int& band, Image& reverseProjRowColImage, Image& forwardProjRowColImage,
                             Image& sourceImage, const int& source_band);
*/
      void registered_tc_copy(const int& band, Image& sourceImage, const int& tc_band, Image& maskImage);
      void registered_ndvi_copy(const int& band, Image& sourceImage, Image& maskImage);
      void registered_feature_copy(const int& process_flag, const int& band, 
                                   Image& sourceImage, const int& source_band, Image& maskImage);
      void registered_mask_copy(Image& sourceImage);
#ifndef GDAL
      void clear_data();
      void close_and_reopen();
#endif
      void flush_data();
      void close();

      void print_info();

    protected:

    private:
      void initialize(); // Called by constructors
      void read_data();
      void write_data();
#ifdef GDAL
      bool geo_copy(const string& file_name, const Image& baseImage, const int& copy_nbands, 
                    const GDALDataType& copy_data_type, const string& copy_driver_description);
#else
      bool geo_copy(const string& file_name, const Image& baseImage, const int& copy_nbands, 
                    RHSEGDType& copy_dtype);
#endif
      int make_histo_eq_map(vector<double>& heap, int& heap_size, vector<double>& histo_eq_map);
      void make_histo_lookup(vector<double>& histo_eq_map, const int& min_histo_index, const double& histo_scale,
                             const double& histo_offset, vector<unsigned char>& histo_lookup);
      void interpolate(const double& continuity_ratio);

   // GDAL and image data file information
      bool gdal_flag;              // True if GDAL valid and associated with image data file
      string image_file;           // Image data file name
#ifdef GDAL
      GDALDataset *imageDataset;   // Pointer to GDAL data set
      GDALDriver  *driver;         // Pointer to GDAL driver for image data file access
      string driver_description;   // GDAL driver description
#endif
      fstream data_fs;

   // Image data information
      bool info_flag;              // True if image information is valid
      int  ncols, nrows, nslices, nbands;   // Number of columns, rows, slices and bands in the image data file
      RHSEGDType dtype;            // RHSEG style data type.
#ifdef GDAL
      GDALDataType data_type;      // GDAL data type of the image data file (complex data not supported)
#endif
      int    io_nrows;             // Number of rows for image data I/O
      int    io_nrows_step;        // Step size for incrementing rows offset for image I/O
      int    io_nrows_overlap;     // Overlap in image I/O when in read only or read and write mode (no overlap in write only mode)
      int    io_nrows_offset;      // Number of rows offset for image I/O (-1 if invalid)
      bool   read_flag;		   // true => read_only
      bool   write_flag;           // true => write_only
      int    io_band;              // In write_only mode, band being written (ignored in read_only mode - all bands read at once)
      vector<double> no_data_value;       // Data values for each band that flag invalid data elements.
      vector<bool>   no_data_value_flag;  // True if corresponding no_data_values are valid
      vector<double> min_value, max_value;           // Minimum and Maximum data values for each band
      vector<bool>   min_value_flag, max_value_flag; // True if corresponding minimum and maximum values are valid

   // Image data geotransform information
      bool geotransform_flag;      // True if geotransform information is valid
      double X_offset, Y_offset, X_gsd, Y_gsd; //  X and Y offset and ground sampling distance
#ifdef GDAL
      string   projection_type;    // Image projection type - empty string if invalid
   // imageGeoTransform[0] = X_offset; imageGeoTransform[1] = X_gsd; imageGeoTransform[2] = 0;
   // imageGeoTransform[3] = Y_offset; imageGeoTransform[4] = 0;     imageGeoTransform[5] = Y_gsd;
      double imageGeoTransform[6]; // GDAL GeoTransform vector
#endif
   // Permissible data arrays. Complex not supported.
      bool data_flag;                   // True if image data is valid
      unsigned char      *byte_data;    // GDT_Byte  or dtype = UInt8
      unsigned short int *ushort_data;  // GDT_UInt16 or dtype = UInt16
      unsigned int       *uint_data;    // GDT_Uint32 or dtype = UInt32
      float              *float_data;   // GDT_Float32 or dtype = Float32
#ifdef GDAL
      short int          *short_data;   // GDT_Int16
      int                *int_data;     // GDT_Int32
      double             *double_data;  // GDT_Float64
#endif

   // Other member variables
      int    red_display_band, green_display_band, blue_display_band;
      vector<bool>   histo_eq_map_flag;
      vector<double> histo_scale;
      vector<double> histo_offset;
      vector< vector<double> > histo_eq_map;
      vector< vector<unsigned char> > histo_lookup;
      vector<ColorMap> colormap;
      bool             colormap_flag;  // True if colormap is valid, false otherwise.
  };

} // CommonTilton

#endif /* IMAGE_H */
